@@ -1,11 +1,22 @@
+"""
+
+Programmer: Ritam Guha
+Date of Development: 6/10/2020
+
+"""
+
 import numpy as np
 import time
 import matplotlib.pyplot as plt
+
 from sklearn.model_selection import train_test_split
-from Py_FS.wrapper.nature_inspired._utilities import Solution, Data, initialize, sort_agents, display, compute_accuracy
 from sklearn import datasets
 
-def GA(num_agents, max_iter, train_data, train_label, obj_function=compute_accuracy, prob_cross=0.4, prob_mut=0.3):
+from Py_FS.wrapper.nature_inspired._utilities import Solution, Data, initialize, sort_agents, display, compute_fitness, compute_accuracy
+# from _utilities import Solution, Data, initialize, sort_agents, display, compute_fitness, compute_accuracy
+
+
+def GA(num_agents, max_iter, train_data, train_label, obj_function=compute_fitness, prob_cross=0.4, prob_mut=0.3, save_conv_graph=False):
 
     # Genetic Algorithm
     ############################### Parameters ####################################
@@ -17,17 +28,23 @@ def GA(num_agents, max_iter, train_data, train_label, obj_function=compute_accur
     #   obj_function: the function to maximize while doing feature selection      #
     #   prob_cross: probability of crossover                                      #
     #   prob_mut: probability of mutation                                         #
+    #   save_conv_graph: boolean value for saving convergence graph               #
     #                                                                             #
     ###############################################################################
 
+    short_name = 'GA'
+    agent_name = 'Chromosome'
+    train_data, train_label = np.array(train_data), np.array(train_label)
     num_features = train_data.shape[1]
     cross_limit = 5
 
     # initialize chromosomes and Leader (the agent with the max fitness)
     chromosomes = initialize(num_agents, num_features)
     fitness = np.zeros(num_agents)
+    accuracy = np.zeros(num_agents)
     Leader_agent = np.zeros((1, num_features))
     Leader_fitness = float("-inf")
+    Leader_accuracy = float("-inf")
 
     # initialize convergence curves
     convergence_curve = {}
@@ -38,7 +55,7 @@ def GA(num_agents, max_iter, train_data, train_label, obj_function=compute_accur
     data = Data()
     data.train_X, data.val_X, data.train_Y, data.val_Y = train_test_split(train_data, train_label, stratify=train_label, test_size=0.2)
 
-    # create a Solution object
+    # create a solution object
     solution = Solution()
     solution.num_agents = num_agents
     solution.max_iter = max_iter
@@ -62,12 +79,25 @@ def GA(num_agents, max_iter, train_data, train_label, obj_function=compute_accur
 
         # update final information
         chromosomes, fitness = sort_agents(chromosomes, obj_function, data)
-        display(chromosomes, fitness)
-        Leader_agent = chromosomes[0].copy()
-        Leader_fitness = fitness[0].copy()
+        display(chromosomes, fitness, agent_name)
+        if fitness[0]>Leader_fitness:
+            Leader_agent = chromosomes[0].copy()
+            Leader_fitness = fitness[0].copy()
         convergence_curve['fitness'][iter_no] = Leader_fitness
         convergence_curve['feature_count'][iter_no] = int(np.sum(Leader_agent))
 
+    # compute final accuracy
+    Leader_agent, Leader_accuracy = sort_agents(Leader_agent, compute_accuracy, data)
+    chromosomes, accuracy = sort_agents(chromosomes, compute_accuracy, data)
+
+    print('\n================================================================================')
+    print('                                    Final Result                                  ')
+    print('================================================================================\n')
+    print('Leader ' + agent_name + ' Dimension : {}'.format(int(np.sum(Leader_agent))))
+    print('Leader ' + agent_name + ' Fitness : {}'.format(Leader_fitness))
+    print('Leader ' + agent_name + ' Classification Accuracy : {}'.format(Leader_accuracy))
+    print('\n================================================================================\n')
+
     # stop timer
     end_time = time.time()
     exec_time = end_time - start_time
@@ -87,15 +117,19 @@ def GA(num_agents, max_iter, train_data, train_label, obj_function=compute_accur
     axes[1].set_xlabel('Iteration')
     axes[1].set_ylabel('Number of Selected Features')
     axes[1].plot(iters, convergence_curve['feature_count'])
-
+    
+    if(save_conv_graph):
+        plt.savefig('convergence_graph_'+ short_name + '.jpg')
     plt.show()
 
     # update attributes of solution
     solution.best_agent = Leader_agent
     solution.best_fitness = Leader_fitness
+    solution.best_accuracy = Leader_accuracy
     solution.convergence_curve = convergence_curve
     solution.final_population = chromosomes
     solution.final_fitness = fitness
+    solution.final_accuracy = accuracy
     solution.execution_time = exec_time
 
     return solution
@@ -128,6 +162,7 @@ def mutation(chromosome, prob_mut):
 
 
 def roulette_wheel(fitness):
+    # Perform roulette wheel selection
     maximum = sum([f for f in fitness])
     selection_probs = [f/maximum for f in fitness]
     return np.random.choice(len(fitness), p=selection_probs)
@@ -137,6 +172,7 @@ def cross_mut(chromosomes, fitness, obj_function, data, prob_cross, cross_limit,
     # perform crossover, mutation and replacement
     count = 0
     num_agents = chromosomes.shape[0]
+    train_X, val_X, train_Y, val_Y = data.train_X, data.val_X, data.train_Y, data.val_Y
     print('Crossover-Mutation phase starting....')
 
     while(count<cross_limit):
@@ -148,8 +184,8 @@ def cross_mut(chromosomes, fitness, obj_function, data, prob_cross, cross_limit,
             child_1, child_2 = crossover(chromosomes[id_1], chromosomes[id_2], prob_cross)
             child_1 = mutation(child_1, prob_mut)
             child_2 = mutation(child_2, prob_mut)
-            fitness_1 = obj_function(child_1, data)
-            fitness_2 = obj_function(child_2, data)
+            fitness_1 = obj_function(child_1, train_X, val_X, train_Y, val_Y)
+            fitness_2 = obj_function(child_2, train_X, val_X, train_Y, val_Y)
 
             if(fitness_1 < fitness_2):
                 temp = child_1, fitness_1
@@ -177,18 +213,12 @@ def cross_mut(chromosomes, fitness, obj_function, data, prob_cross, cross_limit,
             print('Restarting crossover....\n')
 
 
+
+
+
 ############# for testing purpose ################
 
 if __name__ == '__main__':
     iris = datasets.load_iris()
-    GA(10, 20, iris.data, iris.target, compute_accuracy)
-    # a = np.array([1, 0, 0, 1, 1])
-    # b = np.array([0, 1, 1, 0, 0])
-    # c, d = crossover(a, b, 0.4)
-    # print(a)
-    # print(b)
-
-    # print(c)
-    # print(d)
-
+    GA(10, 20, iris.data, iris.target, save_conv_graph=True)
 ############# for testing purpose ################
